﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.AI;

/// <summary>
/// Clase que se encarga de gestionar entidades en la escena
/// </summary>
public class ManejadorDeEscena3D : MonoBehaviour {

    private static ManejadorDeEscena3D reference;

    public ARTrackedObject arScene;                 // escena

    private GameObject generadoresDeEnemigos;       // conjunto de todos los generadores de enemigos de la escena
    private GameObject capsulas;                    // conjunto de todas las cápsulas de la escena
    private GameObject cajas;                       // conjunto de todas las cajas de la escena

    //public PanelDebug panelDebug;   // JPS Debug 27/11/2017

    #region prefabs
    public ParticleSystem particulasImpacto;        // prototipo del sistema de partículas del impacto de un disparo
    public GeneradorDeEntidades generadorCaja1; //
    public Capsula capsula;                         // prototipo de una cápsula
    public Caja caja;                               // prototipo de una caja
    public Enemigo enemigo;                         // prototipo de un enemigo
    public GeneradorDeEnemigos generadorEnemigos;   // prototipo de un punto de generación de enemigos
    public PowerupX2 powerupX2;                     // prototipo de un powerup multiplicador de daño por 2
    #endregion


    private void Start()
    {
        reference = this;

        generadoresDeEnemigos = arScene.transform.Find("GeneradoresDeEnemigos").gameObject;
        capsulas = arScene.transform.Find("Capsulas").gameObject;
        cajas = arScene.transform.Find("Cajas").gameObject;
    }

    /// <summary>
    /// Genera una entidad a partir de un prototipo
    /// </summary>
    /// <param name="p_prefab">prototipo de la entidad a generar</param>
    /// <param name="p_posicion">posición de la entidad a generar</param>
    /// <param name="p_rotacion">rotación de la entidad a generar</param>
    /// <param name="p_escala">escala de la entidad a generar</param>
    /// <returns>La entidad que se ha generado</returns>
    public GameObject GenerarEntidad(GameObject p_prefab, Vector3 p_posicion, Quaternion p_rotacion, Vector3 p_escala)
    {
        GameObject go = Instantiate(p_prefab, arScene.transform, true);
        go.transform.position = p_posicion;
        go.transform.rotation = p_rotacion;
        go.transform.localScale = p_escala;
        return go;
    }

    public EntidadMapa GenerarEntidad(EntidadMapa p_prefab, Vector3 p_posicion, Quaternion p_rotacion, Vector3 p_escala)
    {
        EntidadMapa go = Instantiate(p_prefab, arScene.transform, true);
        go.transform.position = p_posicion;
        go.transform.rotation = p_rotacion;
        go.transform.localScale = p_escala;
        return go;
    }

    
    public static ManejadorDeEscena3D getReference()
    {
        return reference;
    }

    public static ParticleSystem CrearParticulasImpacto(Vector3 p_Posicion, Quaternion p_Rotacion)
    {
        if (reference != null)
            return Instantiate(reference.particulasImpacto, p_Posicion, p_Rotacion, reference.arScene.transform);
        else
            return null;
    }

    public static Capsula CrearCapsula(Vector3 p_Posicion, float p_RotacionY)
    {
        if (reference != null && reference.capsulas != null)
            return Instantiate(reference.capsula, p_Posicion, Quaternion.Euler(0f, p_RotacionY, 0f), reference.capsulas.transform);
        else
            return null;
    }

    public static PowerupX2 CrearPowerupX2(Vector3 p_Posicion)
    {
        if (reference != null)
            return Instantiate(reference.powerupX2, p_Posicion, Quaternion.Euler(0f, 0f, 0f), reference.arScene.transform);
        else
            return null;
    }

    public static Caja CrearCaja(Vector3 p_Posicion, float p_RotacionY)
    {
        if (reference != null && reference.cajas != null)
            //return Instantiate(reference.caja, p_Posicion + new Vector3(0f, .015f, 0f), Quaternion.Euler(0f, p_RotacionY, 0f), reference.arScene.transform);
            return Instantiate(reference.caja, p_Posicion + new Vector3(0f, .015f, 0f), Quaternion.Euler(0f, p_RotacionY, 0f), reference.cajas.transform);
        else
            return null;
    }

    public static GeneradorDeEnemigos CrearGeneradorDeEnemigos(Vector3 p_Posicion, float p_RotacionY, int p_PoblacionMaxima, int p_GeneracionesMaximas, float p_TiempoRecarga)
    {
        if (reference != null)
        {
            GeneradorDeEnemigos ge = Instantiate(reference.generadorEnemigos, p_Posicion, Quaternion.Euler(0f, p_RotacionY, 0f), reference.generadoresDeEnemigos.transform);
            ge.maxPoblacion = p_PoblacionMaxima;
            ge.maxGeneraciones = p_GeneracionesMaximas;
            ge.tiempoRecarga = p_TiempoRecarga;
            return ge;
        }
        else
            return null;
    }

    public static Enemigo CrearEnemigo(Vector3 p_Posicion, float p_RotacionY, Transform p_Parent)
    {
        if (reference != null) {

            Vector3 targetPoint = new Vector3(0f,0f,0f);

            if (reference.capsulas != null)
            {
                int nCapsulas = reference.capsulas.transform.childCount;
                //reference.panelDebug.addTextoDerecha("ncapsulas: " + nCapsulas.ToString()); // DEBUG JPS 28/11/2017
                if (nCapsulas > 0)
                {
                    int index = Random.Range(0, nCapsulas);
                    targetPoint = reference.capsulas.transform.GetChild(index).transform.position;
                    //reference.panelDebug.addTextoIzquierda("capsula index: " + index.ToString()); // DEBUG JPS 28/11/2017
                }
            }
            //reference.panelDebug.addTextoIzquierda("target: " + targetPoint.ToString()); // DEBUG JPS 28/11/2017
            //Debug.Log(string.Format("targetPoint: {0}, {1}, {2}", targetPoint.x, targetPoint.y, targetPoint.z));
            Enemigo en = Instantiate(reference.enemigo, p_Posicion, Quaternion.Euler(0f, p_RotacionY, 0f), p_Parent);
            if (en != null)
            {
                
                en.target = targetPoint;
                

                switch (Preferencias.Dificultad)
                {
                    case "Fácil":

                        en.setNivelFacil();
                        break;

                    case "Normal":

                        en.setNivelNormal();
                        break;

                    case "Difícil":

                        en.setNivelDificil();
                        break;
                }
            } else
            {
                PanelDebug.staticAddTextoDerecha("enemy null");
            }
            //NavMeshAgent nav = en.GetComponent<NavMeshAgent>(); // JPS DEBUG 28/11/2017
            //reference.panelDebug.addTextoIzquierda("Enemigo (" + nav.destination.ToString() + ")"); // JPS DEBUG 27/11/2017
            return en;
        }
        else
            return null;
    }

    /// <summary>
    /// Establece el estado de pausa de los elementos de la escena
    /// </summary>
    /// <param name="p_EstadoPausa">Estado de pausa: true si se pone pausa, false si se reanuda</param>
    public static void enPausa(bool p_EstadoPausa)
    {
        Partida.enPausa = p_EstadoPausa;
        Time.timeScale = p_EstadoPausa ? 0 : 1; // detiene/reanuda la física
        if (reference.generadoresDeEnemigos!=null)
        {
            int i = 0;
            while (i < reference.generadoresDeEnemigos.transform.childCount)
            {
                GeneradorDeEnemigos ge = reference.generadoresDeEnemigos.transform.GetChild(i).gameObject.GetComponent<GeneradorDeEnemigos>();
                if (ge != null)
                    ge.enPausa(p_EstadoPausa);
                i++;
            }
        }
    }

    /// <summary>
    /// Devuelve una lista con todas las cápsulas
    /// </summary>
    /// <returns>Lista de cápsulas de la escena</returns>
    public static List<Transform> getCapsulas()
    {
        int nChilds = getReference().capsulas.transform.childCount;
        List<Transform> childs= new List<Transform>();
        for (int i = 0; i < nChilds; i++)
        {
            Transform child = getReference().capsulas.transform.GetChild(i);
            childs.Add(child);
        }
        return childs;
    }

    private void Update()
    {
        // si no quedan cápsulas en la escena avisa a GamePlay para actuar en consecuencia
        if (reference.capsulas.transform.childCount == 0)
            GamePlay.OnCapsulasDestruidas();
    }
}
